package de.uka.ipd.algorithms;

import de.uka.ipd.algorithms.cfg.CFG;
import de.uka.ipd.algorithms.cfg.CFGNode;

/**
 * Implementation of the lengauerTarjan algorithm. This version annotates
 * the cfg-node labels with the calculated semi and immediate dominators.
 */
public class LengauerTarjan {
	/** class containing additional info about nodes in the cfg */
	private class NodeInfo {
		public CFGNode  node;            
		public int      dfsNumber = -1;  /**< depth first search number */
		public int      parent;          /**< dfsNumber of parent in the dfs-tree */
		public int      sdom;            /**< dfsNumber of semidominator */
		public int      idom;            /**< dfsNumber of immediate dominator */
		public NodeInfo bucket;          /**< linked list of nodes with same semidominator */ 
		public int      ancestor  = -1;  /**< last ancestor of the node which
		                                      has been processed for finding
		                                      smallest semidom */
		public int      smallest_sdom = -1; /**< ancestor with smallest sdom
		                                         (smallest one until the one in the
		                                          ancestor field) */
	}

	/** maps node indices to dfs numbers */
	private NodeInfo[] nodeInfos;

	/** returns additional info for CFG nodes, indexed by dfs numbers (not cfg node number!) */
	private NodeInfo[] realNodes;
	/** next free dfs number */
	private int        nextDfsNumber;

	/**
	 * do a depth first search and construct nodeinfos for each node in the CFG.
	 * The nodeinfos are preinitialized with the depth-first-search-number
	 * (dfsNumber) of the node and it's parent in the depth-first spanning tree.
	 */
	private void depthFirstSearch(CFGNode node, int parentDfs) {
		int index = node.getIndex();
		int dfsNumber = nextDfsNumber;
		++nextDfsNumber;

		NodeInfo info  = new NodeInfo();
		info.node      = node;
		info.dfsNumber = dfsNumber;
		info.parent    = parentDfs;
		nodeInfos[dfsNumber] = info;

		assert realNodes[index] == null;
		realNodes[index] = info;

		node.setLabel(node.getLabel() + "(" + info.dfsNumber + ")");

		for (CFGNode succ : node.getSuccs()) {
			int succIndex = succ.getIndex();
			/* already visited? */
			if (realNodes[succIndex] != null)
				continue;

			depthFirstSearch(succ, dfsNumber);
		}
	}

	private int getAncestorWithLowestSemi(int dfsNumber) {
		NodeInfo info         = nodeInfos[dfsNumber];
		int      ancestor     = info.ancestor;
		NodeInfo ancestorInfo = nodeInfos[ancestor];
		
		/* parent not processed yet? return smallest sdom so far */
		if (ancestorInfo.ancestor < 0)
			return info.smallest_sdom;
		
		/* else calculate lowest semi and do path compression */
		int lowestSemi = getAncestorWithLowestSemi(ancestor);
		info.ancestor  = ancestorInfo.ancestor;
		if (nodeInfos[lowestSemi].sdom < nodeInfos[info.smallest_sdom].sdom) {
			info.smallest_sdom = lowestSemi;
		}
		
		return lowestSemi;
	}

	private void calculateSemiDoms() {
		/* look at all nodes in reverse order of dfs-Numbers */
		for (int i = nodeInfos.length - 1; i >= 1; --i) {
			NodeInfo info   = nodeInfos[i];
			CFGNode  node   = info.node;
			int      parent = info.parent;
			
			/* determine semidominator */
			int sdom = parent;
			for (CFGNode pred : node.getPreds()) {
				int tempSemi;
				int predDfsNumber = realNodes[pred.getIndex()].dfsNumber;

				if (predDfsNumber < i) {
					tempSemi = predDfsNumber;
				} else {
					int ancestorLowestSemi = getAncestorWithLowestSemi(predDfsNumber);
					tempSemi = nodeInfos[ancestorLowestSemi].sdom;
				}
				
				/* better than our previously found sdom candidates? */
				if (tempSemi < sdom)
					sdom = tempSemi;
			}
			info.sdom = sdom;

			CFGNode semiNode = nodeInfos[sdom].node;
			node.setLabel(node.getLabel() + "\nsdom = " + semiNode.getName());

			/* initialize ancestor (we processed the node and know it's
			 * semidominator now. Our dfs-Number will be bigger than the ones
			 * of all further nodes we process, so we will be interested in
			 * this path when going up and looking for smallest semidominator
			 * node with higher dfs-Number */
			info.ancestor      = parent;
			info.smallest_sdom = i; 
			
			/* add node to list of semidominated nodes */
			NodeInfo semiDomNodeInfo = nodeInfos[sdom];
			info.bucket              = semiDomNodeInfo.bucket;
			semiDomNodeInfo.bucket   = info;

			NodeInfo parentInfo = nodeInfos[parent];
			while (parentInfo.bucket != null) {
				NodeInfo b = parentInfo.bucket;
				
				/* remove b from bucket list */
				parentInfo.bucket = b.bucket;
				b.bucket          = null;
				
				int y = getAncestorWithLowestSemi(b.dfsNumber);
				
				if (nodeInfos[y].sdom < b.sdom) {
					b.idom = y;
				} else {
					b.idom = nodeInfos[parent].dfsNumber;
				}
			}
		}

		for (int i = 1; i < nodeInfos.length; ++i) {
			NodeInfo info = nodeInfos[i];
			CFGNode  node = info.node;
			
			if (info.idom != info.sdom) {
				info.idom = nodeInfos[info.idom].idom;
			}

			node.setLabel(node.getLabel() + "\nidom = " + nodeInfos[info.idom].node.getName());
		}
	}

	/* private constructor so others can't instantiate the LengauerTarjan class */
	private LengauerTarjan() {
	}

	public static void DoLengauerTarjan(CFG cfg) {
		LengauerTarjan instance = new LengauerTarjan();

		instance.nodeInfos     = new NodeInfo[cfg.nodeCount()];
		instance.realNodes     = new NodeInfo[cfg.nodeCount()];
		instance.nextDfsNumber = 0;
		instance.depthFirstSearch(cfg.getStart(), -1);
		instance.calculateSemiDoms();
	}
}
